(** Operations on CDuce values *)

val print: Format.formatter -> {{ Any }} -> unit
  (** Pretty-print a value as in the toplevel. *)

val to_string: {{ Any }} -> string
  (** Same as [print] but returns the value in a string. *)

module Utf8 : sig
  type t = utf8
      
  val to_string_utf8: t -> string
  val of_string_utf8: string -> t

  val to_string_latin1: t -> string
  val of_string_latin1: string -> t
end

module Namespaces : sig
  (** Operations on namespace bindings. *)

  type t = (string * string) list
      (** The type for namespace bindings: associative lists
	  from prefixes to namespace URIs. All strings are UTF-8. *)

  val get: {{ <(Any) (Any)>Any }} -> t
    (** [get v] extracts the namespace bindings from the value [v]
	or raise [Not_found] if [v] has no attached namespace bindings. *)
    
  val set: {{ <(Any) (Any)>Any }} -> t ->  {{ <(Any) (Any)>Any }}
    (** [set v ns] returns a value equal to [v] except for the attached 
	namespace bindings which are set to [ns]. *)
end

module Load : sig
  (** Importing XML documents. *)

  (** This module provides an abstract type of XML loaders.
      An XML parser can interact with such an object to produce a CDuce value.
      After creating the loader with the [make] function, the XML parser
      must call [start_elem], [end_elem] and [text] for each
      corresponding event found in the input document. The loader
      object is responsible for XML Namespaces resolution.
      All the strings must be encoded in UTF-8. *)

  type t
    (** The type of loader objects. *)
  type anyxml = {{ <_ ..>[ (Char | anyxml)* ] }}
      (** The type for the values produced by the loader. *)
      
  exception Error of string
    
  val make: ?ns:bool -> unit -> t
    (** [make ~ns ()] creates a fresh loader. If [ns] is [true],
	the loader will attach a namespace bindings to each XML element. *)
  val get: t -> anyxml
    (** [get l] is to be called when the whole XML document has been parsed.
	It returns the root element of the loaded XML documents, or
	raises [Error] if the document is not well-formed. *)

  val start_elem: t -> string -> (string * string) list -> unit
    (** [start_elem l tag attr] is to be called when the XML parser
	encounters the opening markup for an XML element
	with specified tag and attributes.
	The function raises [Error] if the same attribute label
	appears several times. *)

  val end_elem: t -> 'a -> unit
    (** [end_element l x] is to be called when the XML parser
	encounters the closing markup for an XML element. *)
 
  val text: t -> string -> unit
    (** [text l txt] is to be called when the XML parser
	encounters some textual data (PCDATA). It is legal
	to call [text] several times in a row. *)
end

module Print : sig
  (** Exporting XML documents. *)

  exception Error of string * {{ Any }}

  val print_xml: 
    markup:(string -> unit) -> ?data:(string -> unit) -> 
    ?ns:Namespaces.t -> 
    Load.anyxml ->
    unit
      (** [print ~markup ~data ~ns v] produces textual
	  content corresponding to the XML representation
	  of the value [v]. The function [markup] is called for markup strings
	  (tags, attribute names, ...). The function [data] is called 
	  for textual data (attribute value, PCDATA). If [data] is not 
	  provided, the function [markup] is used instead.
	  Strings passed to [markup] and [data] are encoded in UTF-8. 

	  The function attach bindings for XML Namespaces used
	  in the tree to the root element.
	  The optional argument [ns] can be used to provide
	  hints to choose the prefixes.
	  
	  If the value does not represent a valid XML tree,
	  the exception [Error] is raised.
      *)
end
