let cu = Types.CompUnit.mk (Encodings.Utf8.mk "OCAML")
let () = Types.CompUnit.enter cu; Types.caml_mode := true
let init = ref []
let serialize = ref []

module Mk(X : Custom.T) = struct
  module A = Custom.Array(X)
  
  type pchunk = { mutable nb : int; mutable lst : X.t list }
  let put c x = 
    let i = c.nb in
    c.nb <- succ i;
    c.lst <- x::c.lst;
    i
    
  let init () = { nb = 0; lst = [] }
  let serialize s c = Serialize.Put.array X.serialize s 
		      (Array.of_list (List.rev c.lst))

  type 'a entry = Serialized of X.t | Computed of 'a
  type 'a chunk = 'a entry array

  let deserialize s = 
    Serialize.Get.array (fun s -> Serialized (X.deserialize s)) s

  let get f a i =
    match a.(i) with
      | Serialized x ->
	  let x = f x in
	  a.(i) <- Computed x;
	  x
      | Computed x ->
	  x
end

module PM = Mk(Custom.Pair(Types)(Custom.List(Patterns.Node)))
module CONST = Mk(Types.Const)
module LAB = Mk(Ident.LabelPool)
module T = Mk(Types)
module LABA = Mk(Custom.Array(Ident.LabelPool))
module TAG = Mk(Atoms.V)
module TAGA = Mk(Custom.Array(Custom.Pair(Atoms.V)(Custom.Int)))

module P = struct
  type chunk =
      { pm : PM.pchunk;
	cst : CONST.pchunk;
	lab : LAB.pchunk;
	typ : T.pchunk;
 	laba : LABA.pchunk;
	tag : TAG.pchunk;
	taga : TAGA.pchunk;
     }

  let init () = 
    { pm = PM.init ();
      cst = CONST.init ();
      lab = LAB.init ();
      typ = T.init ();
      laba = LABA.init ();
      tag = TAG.init ();
      taga = TAGA.init ();
    }

  let serialize s c =
    PM.serialize s c.pm;
    CONST.serialize s c.cst;
    LAB.serialize s c.lab;
    T.serialize s c.typ;
    LABA.serialize s c.laba;
    TAG.serialize s c.tag;
    TAGA.serialize s c.taga;
    ()

  let pm c = PM.put c.pm
  let const c = CONST.put c.cst
  let label c = LAB.put c.lab
  let typ c = T.put c.typ
  let label_array c = LABA.put c.laba
  let tag c = TAG.put c.tag
  let tag_array c = TAGA.put c.taga

  let mk c =
    let s = Serialize.Put.run serialize c in
    ignore (Types.CompUnit.close_serialize ());
    s

end

module G = struct
  type chunk =
      { pm :
	  (Patterns.Compile.dispatcher * int Patterns.Compile.rhs array)
	  PM.chunk;
	cst : Value.t CONST.chunk;
	lab : Ident.label LAB.chunk;
	typ : Types.t T.chunk;
	laba : Ident.label array LABA.chunk;
	tag : Value.t TAG.chunk;
	taga : int Atoms.map TAGA.chunk;
      }

  let deserialize s =
    let pm = PM.deserialize s in
    let cst = CONST.deserialize s in
    let lab = LAB.deserialize s in
    let typ = T.deserialize s in
    let laba = LABA.deserialize s in
    let tag = TAG.deserialize s in
    let taga = TAGA.deserialize s in
    { pm = pm; cst = cst; lab = lab; typ = typ; laba = laba; tag = tag;
      taga = taga }

  let mk s = 
    Types.clear_deserialize_table ();
    Serialize.Get.run deserialize s

  let mk_pm (t,brs) =
    let brs = Array.to_list (Array.mapi (fun i x -> (x,i)) (Array.of_list brs))
    in
    Patterns.Compile.make_branches t brs

  let pm chunk i v =
    let (d,rhs) = PM.get mk_pm chunk.pm i in
    let (code,bindings) = Run_dispatch.run_dispatcher d v in
    match rhs.(code) with
      | Patterns.Compile.Fail -> (-1,[||])
      | Patterns.Compile.Match (bind,i) ->
	  i,
	  Array.map
	    (fun (_,i) -> if (i == -1) then v else bindings.(i))
	    (Array.of_list bind)


  let const chunk i =
    CONST.get Value.const chunk.cst i
	  
  let remove_label chunk i v =
    Value.remove_field (LAB.get (fun x -> x) chunk.lab i) v
  let get_field chunk i v =
    Value.get_field v (LAB.get (fun x -> x) chunk.lab i)

  let typ chunk i =
    T.get (fun x -> x) chunk.typ i

  let check chunk t0 t v =
    let t0 = typ chunk t0 and t = typ chunk t in
    match Explain.explain t0 t v with
    | None -> v
    | Some p -> failwith (Explain.to_string p)
  

  let record chunk i vs =
    Value.mk_record (LABA.get (fun x -> x) chunk.laba i) vs

  let constr_const chunk i =
    TAG.get (fun x -> Value.Atom x) chunk.tag i

  let constr chunk i vs =
    Value.ocaml2cduce_constr (constr_const chunk i) vs

  let dconstr chunk i v =
    Value.cduce2ocaml_constr (TAGA.get
		      (fun x ->
			 let x = Array.map (fun (t,i) ->
					      Atoms.atom t, i) x in
			 Atoms.mk_map (Array.to_list x))
		      chunk.taga i) v
end

